/*
 * judycam.c - cam memory for hashblist module based on judy trees
 *
 *  TBD. index not in net byte order
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>

#include <Judy.h>
int debug;

/* associative array search using JudyL arrays
*/

static Pvoid_t Parray=(Pvoid_t) NULL;     // JudyL array

/* promote to monconfig.c or utils.c
*/
static void convertNetMask(char *, unsigned int *, unsigned int *);
 
/* pass in value as static structure, malloc and copy
*/
static void insertJudyCAM(Word_t index, Word_t offset, char *ts);
/* make sure ts gets passed in as string
*/
static int sanityCheckJudyCAM(Word_t baseaddr, Word_t netmask, char *ts);

/* runtime search - replaces hash lookups
*/
static int searchJudyCAM(Word_t index);

/* judy value */
struct bnode {
	Word_t netmask;		/* the offset kept as host-order bitmask */
	unsigned short port;	/* port value */
	unsigned short wildcard;   /* true if port wildcard */
	char *tag;		/* string tag (which file) */
};

/* insert index, offset in Judy CAM.  
 *
 * ts - string version of net/mask
 */
static
void 
insertJudyCAM(Word_t index, Word_t offset, char *ts)
{
	PPvoid_t PPvalue;
	int rc;

	/* note: this function won't allow any overlap
	*/
	rc = sanityCheckJudyCAM(index, offset, ts);
	if (rc == 0)
		return;

	/**************************************************/
	/* insert index, offset */
	JLI(PPvalue, Parray, index);	  // insert of index, returns value

	/* malloc failure 
	*/
	if (PPvalue == PJERR) {
		fprintf(stderr, "malloc failure with judy trees\n");
		exit(1);
	}

	*((Word_t *)PPvalue) 
			= offset;    // store of value

	printf("insert: index, size %lx %lx \n", index, *((Word_t *)PPvalue)); 
}

/* sanityCheckJudyCAM 
 *
 * basic idea:  we don't allow overlap as it is confusing
 *	as to what it might mean in terms of handling it.
 * rules: illegal if
 *	1. see if the base address matches a previous base address
 *	2. see if the base address is within a range.
 *		note: 1/2 are basically the same
 *	3. see if the base and the offset (interpreted as an address)
 *		have different predecessors.
 *
 *	returns: 1  ok
 *		 0  not ok
 */
static
int
sanityCheckJudyCAM(Word_t baseaddr, Word_t netmask, char *ts) 
{
	struct in_addr in; 
	Word_t indexstart = baseaddr;
	Word_t foundmask;
	Word_t netmaskstart;
	PPvoid_t PPvalue_base;
	PPvoid_t PPvalue_offset;
	int checkoffset = 0;

	/* lookup baseaddr with predecessor check
	 * note: indexstart is modified if PPvalue is not null
	*/
	JLL(PPvalue_base, Parray, indexstart);	  // last index =< index passed in
	/* check baseaddr for inside region
	*/
	if (PPvalue_base != NULL) {
		foundmask = *((Word_t *)PPvalue_base);
		if ((indexstart & foundmask)  ==  (baseaddr & foundmask)) {
			in.s_addr = (unsigned int) htonl(indexstart);
			fprintf(stderr, "WARNING: addr (%s) already in previous block with base %s, new entry discarded\n", ts, 
			inet_ntoa(in));
			return(0);
		}
	}
	/* note that it is still possible that we didn't find
	 * the baseaddress (value is NULL), we now lookup the offset
	*/
	netmaskstart = baseaddr | ~netmask;
	JLL(PPvalue_offset, Parray, netmaskstart);	  

	/*
	possibilities:
	1. PPvalue_offset is NULL,  PPvalue_base is NULL
		that's ok
	2. PPvalue_offset is not NULL, PPvalue_base is NULL
		bad as overlap
	3. PPvalue_offset is NULL, PPvalue_base is not NULL
		crazy as impossible
	4. neither are null
		ok if same base (we already know that base does
			not overlap)
	*/
	/* takes care of 1/4 */
	if (PPvalue_offset == PPvalue_base) {
		return(1);
	}
	in.s_addr = (unsigned int) htonl(indexstart);
	fprintf(stderr, "WARNING: discarding (%s) due to overlap with previous net/mask base\n", ts);

	return(0);
}

/*
 * note in real world: index needs to be converted
 * to host byte order
 *
 *	returns:
 *		1 - found it.
 *		0 - didn't find it.
 */
int
searchJudyCAM(Word_t index)
{
	PPvoid_t PPvalue;
	Word_t indexstart = index;
	Word_t offset;
	extern int debug;

	if (debug)
		printf("searching judy cam on %lx\n", index);

	/* no mallocs in here */
	JLL(PPvalue, Parray, indexstart);	  // last index =< index passed in
	/* note: PPvalue may be NULL, indexstart may be modified
	*/

	/* if the returned value is null that means not found.
	 * OR if found, the base+offset is still less than index
	 */
	if (PPvalue == (PPvoid_t) NULL) {
		if (debug)
			printf("NULL, no predecessor: %lx not in any region\n", index);
		return (0);
	}

	offset =  *((Word_t *)PPvalue);

	/* if base + offset < index then index not in region
	*/
	printf("%lx %lx\n", indexstart & offset, index & offset);
	if (((indexstart & offset)  ==  (index & offset))) {
		if (debug)
			printf("%lx is in region base=%lx offset = %lx\n", 
			index, indexstart, offset);
		return(1);
	}
	if (debug)
		printf("%lx is OUTSIDE region base=%lx offset = %lx\n", 
			index, indexstart, offset);
	return(0);
}

/* 
 * convertNetMask() - convert string to host net/mask.
 *
 * ns is network string as either:
 *		1.2.3.4
 *		1.0.0.0/24
 *
 * If no slash appears we assume /32
 *
 * input: ns
 * output:  base/mask as host-order unsigned ints.
 *	careful with this on 64-bit gccs.
 *	note *host-order*, may need conversion to net byte order.
 */
static void 
convertNetMask(char *ns, unsigned int  *base, unsigned int *mask)
{
	char *m;
	int bits;
	unsigned int netmask = 0xffffffff;   /* defaut if no netmask */
	struct in_addr ipaddr;
	int rc;
	char buf[1024];

	/* strchr wants an array, not a pointer
	*/
	snprintf(buf, 1024, ns);

	/* look for slash, if no slash then assume /32 
	*/
	m = strchr(buf, '/');
	if (m) {
		/* get the bits as an unsigned long */
		bits = strtoul(m+1, 0, 0);
		if (bits > 32) {
			fprintf(stderr, "invalid ip address, bad mask in %s\n", ns);
			exit(1);
		}
		if (bits) {
			netmask = 0xffffffff << (32 - bits);	
			/*
			netmask = htonl(netmask);
			*/
		}
		*m = 0;  /* make two strings */
	}

	/* ns (buf) is still a string */ 
	/* netmask is in net byte order */

	/* make sure IP address is valid with inet_aton()
	*/
	rc = inet_aton(buf, &ipaddr);
	if (rc == 0) {
		fprintf(stderr, "invalid ip address %s\n", ns);
		exit(1);
	}

	/* ipaddr and netmask are host byte order */
	*base = (unsigned int) ntohl(ipaddr.s_addr);
	*mask = (unsigned int) netmask; 
#define DEBUG
#ifdef DEBUG
	printf("convert: %lx %lx %lx\n", *base, *mask, *base | ~*mask);
#endif
}


main()
{
	char *ts;
	int rc;

	/* note Word_t is unsigned long */

	unsigned int addr, mask;
	Word_t index;
	Word_t offset;
	Word_t indexstart;


	/*
	ts = "1.0.0.0/24";
	*/
	ts = "1.1.1.1/32";
	convertNetMask(ts,  &addr,  &mask);
	printf("%s->%lx %lx\n", ts, addr, mask);
	index = (unsigned long)addr;
	offset = (unsigned long)mask;
	insertJudyCAM(index, offset, ts);

	ts = "2.1.1.1/32";
	convertNetMask(ts,  &addr, &mask);
	printf("%s->%lx %lx\n", ts, index, offset);
	index = (unsigned long)addr;
	offset = (unsigned long)mask;
	insertJudyCAM(index, offset, ts);

	ts = "3.2.2.2";
	convertNetMask(ts, &addr,  &mask);
	printf("%s->%lx %lx\n", ts, index, offset);
	index = (unsigned long)addr;
	offset = (unsigned long)mask;
	insertJudyCAM(index, offset, ts);

/*
	ts = "1.0.0.0/16";
	convertNetMask(ts, (unsigned int) &index, (unsigned int) &offset);
	printf("%s->%lx %lx\n", ts, index, offset);
	insertJudyCAM(index, offset, ts);
*/
	
	/**************************************************/
	/* search for region */
	ts = "3.2.2.2";
	convertNetMask(ts, &addr, &mask);
	index = (unsigned long)addr;
	offset = (unsigned long)mask;
	rc = searchJudyCAM(index);
	if (rc) {
		printf("judycam search on %s successful\n", ts);
	}

	ts = "2.1.1.1";
	convertNetMask(ts, &addr, &mask);
	index = (unsigned long)addr;
	offset = (unsigned long)mask;
	rc = searchJudyCAM(index);
	if (rc) {
		printf("judycam search on %s successful\n", ts);
	}
}
