/*
 * hashblist.c - given a blacklist of IP addresses X, Y, Z
 *	auto store any packets from them into a tcpdump file.
 *
 * config assumptions:
 *
 *	We need to config and turn on a pcap dump capability
 *	We need to parse a list of IP addresses and put them in a hashfile.
 *
 * syntax:
 *	blist_dumpfile maxpackets /usr/dumps
 *		-> filename is /usr/dumps/blacklist.timestamp.dmp
 *		maxpackets may be 0 meaning infinity.
 *	blist_include tag filename (list of ips and ports)
 *	blist_include tag filename (list of ips and ports)
 *	blist_include tag filename (list of ips and ports)
 *	blist_include tag filename (list of ips and ports)
 *	any number of include files, only one trigger file.
 *
 *	config code turns the trigger on 
 *	there is only one trigger file.
 *
 *	we do our own writes as analyze_ip hands us the packet to check.
 *
 */

/*
 * Copyright (c) 2006 Portland State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse
 *    or promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <syslog.h>
#include <pcap.h>
#include <string.h>
#include <sys/stat.h>
#include <Judy.h>

#include "ourmon.h"
#include "hashblist.h"
#include "sync.h"

#ifndef TRUE
#define TRUE    1
#endif
#ifndef FALSE
#define FALSE   0
#endif

extern int debug;

static struct blist bl;
static Pvoid_t Parray=(Pvoid_t) NULL;     // JudyL array

/* in utils.c */
void convertNetMask(char *, unsigned int *, unsigned int *);

static void 
insertJudyCAM(Word_t index, struct bnode *, char *ts);

/* make sure ts gets passed in as string
*/
int 
sanityCheckJudyCAM(Word_t baseaddr, Word_t netmask, char *ts);

/* runtime search - replaces hash lookups
 * TBD. returns: NULL or bnode pointer
*/
inline
static struct bnode *
searchJudyCAM(Word_t index);

static void parseIp(char *, char *, char *, int, char *);

/* called due to blist_dumpfile config directive.
 * create pcap dump fd aimed at dirname/filename
 * config time code:
*/
void
config_blist_dumpfile(char *dirname, unsigned int pktcount)
{
        char ts[100];
        char filename[2100];
        struct stat statbuf;
        extern  int errno;
	extern pcap_t *pd;	/* in ourmon.c */
	char *dfile = "blacklist";
	struct blist *l = &bl;
	char *fn = filename;

	l->count = l->pktcount = pktcount; /* it can't be wrong as it can't be negative */
	l->match = 0;

        /* check the existence of directory 
	*/
        if( stat(dirname, &statbuf) < 0 ) {
                fprintf(stderr,"blist_dumpfile: stat of directory %s failed: error: %s\n", dirname, strerror(errno));
                exit(1);
        }
        if( S_ISDIR(statbuf.st_mode) == 0 )  {
                fprintf(stderr,"blist_dumpfile: dirname %s is not a directory\n", dirname);
                exit(1);
        }

	/* form the name of the dump file
	*/
        bzero(ts, 100);
        bzero(filename, 2100);  
 
        get_ts(ts);
        if( strlen(ts) == 0 ) {
                fprintf(stderr,"blist_dumpfile: get_ts timestamp call failed\n");
                exit(1);
        }

        /* form the config time base filename, which is used
         * at runtime to construct a dynamically instanced unique filename
        */
        if( (strlen(dirname)+strlen(dfile)) > (2048-1) ){
                fprintf(stderr,"blist_dumpfile: dump file name too long\n");
                exit(1);
        }
  
        /* filename = typename.<timestamp>.dmp
        */
        sprintf(filename,"%s/%s.<%s>.dmp", dirname, dfile, ts);

        if((l->pcapfd = pcap_dump_open( pd, (const char *)fn)) == NULL) {
                fprintf(stderr,
                        "blist_dumpfile pcap open failed: %s\n", pcap_geterr(pd));
                exit (1);
        }
	return;
} 

/* insert entries into the hash list from the blacklist file
 *
 * char *tag - one string tag for all the entries in the
 * char *filename - file with list of IP/ports
*/
void
config_blist(char *tag, char *filename)
{
	FILE *fd;
#define MAXBLISTLINE 1024
	char            linebuf[MAXBLISTLINE];
	int ccount; 
	char *opword;
	char           *result;
	int             lineNumber = 1;
	struct blist *l = &bl;
	char opbuf[1024];
	static int firsttime = 1;
	char *sltag;

	/* init hash array first time called
	 * so we can call this again to add more files with lists
	*/
	if (firsttime) {
		firsttime = 0;
		l->entries = 0;
		/* init the string list/s for either runtime memory of actions
		 * or config time memory of tag strings
		*/
		initString(&l->sl);
		initString(&l->sltags);
	}

	/* store the tag and later put a pointer to it in all
	 * hash entries
	 */
	sltag = storeString(&l->sltags, tag);
	if (sltag == NULL) {
		fprintf(stderr, "blist_include: ? out of malloc space for tag %s\n", tag);
		exit(1);
	}

	/* open the blacklist config file.  
	*/
	fd = fopen(filename, "r");
	if (fd == NULL) {
		fprintf(stderr, "blist_include open of blacklist config file %s failed\n", filename);
		exit(1);
	}

	/* XXX - this code is duplicated in monconfig.c, fix that somehow.
	 *
	 * while not at EOF read a line, parse it
	 */
	while (feof(fd) == 0) {
		result = fgets(linebuf, MAXBLISTLINE, fd);
		if (result == NULL)
			break;

		ccount = strlen(linebuf);
		linebuf[ccount-1] = 0;

		/*
		 * skip comment
		 */
		if (linebuf[0] == '#') {
			lineNumber++;
			continue;
		}
		/*
		 * skip empty lines
		 */
		if (linebuf[0] == '\n' || linebuf[0] == '\r') {
			lineNumber++;
			continue;
		}
		/*
		 * skip whitespace in line
		 */
		while (*result == ' ' || *result == '\t') {
			result++;
		}
		/*
		 * find first token
		 */
		strncpy(opbuf, result, 1024);
		opword = strtok(opbuf, " \t");
		if (opword == NULL || (strlen(opword) == 0)) {
			lineNumber++;
			continue;
		}
		/*
		 * sscanf(result, "%s %s", op, value); printf("op value %s
		 * %s\n", op, value);
		 */
		(void) parseIp(filename, linebuf, result, lineNumber, sltag);
		lineNumber++;
	} 
	fclose(fd);
	return;
}

/*
 * note in real world: index needs to be converted
 * to host byte order
 *
 *	returns:
 *		bnode * - found it.
 *		NULL - didn't find it.
 */
inline struct bnode *
searchJudyCAM(Word_t index)
{
	PPvoid_t PPvalue;
	Word_t indexstart = index;
	struct bnode *bn;

	if (debug)
		printf("SEARCHJUDY: searching judy cam on %lx\n", index);

	/* no mallocs in here */
	JLL(PPvalue, Parray, indexstart);	  // last index =< index passed in
	/* note: PPvalue may be NULL, indexstart may be modified
	*/

	/* if the returned value is null that means not found.
	 * OR if found, the base+offset is still less than index
	 */
	if (PPvalue == (PPvoid_t) NULL) {
		if (debug)
			printf("SEARCHJUDY: NULL, no predecessor: %lx not in any region\n", index);
		return (NULL);
	}

	bn =  (struct bnode *)*((Word_t *)PPvalue);

	/* check to see if target actually within predecessor region
	*/
	if (debug)
		printf("SEARCHJUDY %lx %lx\n", indexstart & bn->netmask, index & bn->netmask);
	if ((((unsigned int)indexstart & bn->netmask)  ==  (
			(unsigned int)index & bn->netmask))) {
		if (debug)
			printf("SEARCHJUDY %lx is in region base=%lx netmask = %lx\n", 
			index, indexstart, bn->netmask);
		return(bn);
	}
	if (debug) {
		printf("SEARCHJUDY %lx is OUTSIDE region base=%lx offset = %lx\n", 
			index, indexstart, bn->netmask);
	}
	return(NULL);
}

static void
parseIp(char *filename, char *linebuf, char *result, int lineNumber, char *sltag)
{
	int rc;
	unsigned int ipaddr;
	struct bnode *n;   
	struct blist *l = &bl;
#define MAXLINE 512
	char dc[MAXLINE];
	char v1[MAXLINE];
	unsigned int wildcard = 0;
	unsigned short port;
	extern int errno;
	unsigned int netmask;

	/* 
 	 * dc is net/mask
	 * v1 is port (or any)
	*/
	sscanf(linebuf, "%s %s", dc, v1);

	/* exits if error else returns host-byte order ipaddr/netmask 
	 * note: does not alter dc
	*/
	convertNetMask(dc, &ipaddr, &netmask);

	/* any is legal wildcard, represented internally by -1 */
	if (strncmp(v1, "any", 3)==0) {
		wildcard = 1;
		port = 0;                 
	} 
	else {
		wildcard = 0;
		/* numeric: decide if port is ok
		*/
		port = (unsigned short) strtol(v1, NULL, 10);
		if (errno == EINVAL) {
			fprintf(stderr, "parseIp: invalid port: %s:%s\n", filename, linebuf);
			exit(1);
		}
		/* wrong due to typing but leave alone just in case
		*/
		if (port > 65535 || port < 0) {
			fprintf(stderr, "parseIp: port out of range (0..65535): %s:%s\n", filename, linebuf);
			exit(1);
		}
	}

	/* search judy tree to see if address exists
	 * issues warning if failure, does not exit
	*/
	rc = sanityCheckJudyCAM((Word_t) ipaddr, 
		(Word_t) netmask, dc);
	if (rc == 0) {
		if (debug)
			printf("SANITYJUDY: sanity check failed on %s\n", dc); 
		return;
	}

	/* ok - the net/mask combo is good to go
	 * as it is not in the judy CAM tree
	*/

	/* note this malloc is a config-time malloc only
	 * goal is: allocate bnode, judy hash allocated in 
	 * insertJudyCAM function. 
	*/
	n = (struct bnode *) malloc(sizeof(struct bnode));
	if (n == NULL) {
		fprintf(stderr, "out of space with malloc in hashblist.c\n");
		exit(1);
	}
	/* fill it in
	*/
	n->tag = sltag;
	n->netmask = netmask;
	n->port = (unsigned short) port;
	n->wildcard = wildcard;
	l->entries += 1;

	insertJudyCAM((Word_t) ipaddr, n, dc);
}

/* runtime: packet analysis routine
 * see if pkt should be shoved into tcpdump file
 */
void
search_blist(struct pcap_pkthdr *h, u_char *p, struct t_entry *te)
{
	struct bnode *bn;
	int match = 0;
	struct blist *l = &bl;

	/* if the src or dst address in the packet is mentioned in the blacklist
	 * if this is the 1st time this has happened, we open a trigger file,
	 * and store any blacklist packets in that particular tcpdump
	*/
	unsigned int ipsrcaddr = te->src.s_addr;
	unsigned int ipdstaddr = te->dst.s_addr;
	unsigned short srcport = te->sport;
	unsigned short dstport = te->dport;
	unsigned short ipproto = te->proto;

	/* match if address is src address in packet AND
	 * either port if port not a wildcard
	 * TBD:
	 * this is a little fuzzy and perhaps should be 
	 * ipsrc and src port or ipdst and dst port.
	*/
	bn = searchJudyCAM((Word_t) ntohl(ipsrcaddr));
	if (bn && (bn->wildcard || 
		((bn->port == te->sport) || (bn->port == te->dport)))) {
		match = 1;
	}
	if (!match) {
		bn = searchJudyCAM((Word_t) ntohl(ipdstaddr));
		if (bn && (bn->wildcard || 
			((bn->port == te->sport) || (bn->port == te->dport)))) {
			match = 1;
		}
	}
	if (debug) {
		if (match)
			printf("BLACKLIST: match %d ip %x:%x s=%d d=%d\n", 
				match, ipsrcaddr, ipdstaddr, srcport, dstport);
	}

#ifdef THREAD
	/* spinlock because of possible file i/o and malloc in string storage
	*/
	spinlock(&locks[LOCK_BLIST]); 
#endif
	if (match) {
		l->match = 1;
		/* write packet via blist->pcapfd, te->h, te->op 
		*/
		if (l->pktcount==0 || l->count--) {
			pcap_dump((u_char *)l->pcapfd, h, p);
		}
		/* memorize what happened
		*/
		{
#define BUFSIZE 1024
			char buf[BUFSIZE];
			char addr1[BUFSIZE];
			int rc;
			rc = sprintf(addr1,"%s", intoa(ipsrcaddr));
			addr1[rc] = 0;
			rc = sprintf(buf, "(%s):%s.%d->%s.%d[%d]", 
				bn->tag, addr1, srcport, 
				intoa(ipdstaddr), dstport, ipproto);
			buf[rc] = 0;
			storeString(&l->sl, buf);
		}
	}
#ifdef THREAD
 	unlock(&locks[LOCK_BLIST]);
#endif
} 

void 
print_blist(FILE *monfd)
{
	struct blist *l = &bl;

	if (l->match) {
		/* elog message if desired
		fprintf(monfd, "elog: IP blacklist storage has occured \n");
		*/
		dumpString(&l->sl, monfd, "elog: IP blacklist event:");
		zeroString(&l->sl);
		initString(&l->sl);
		
		l->match = 0;
	}
	/* flush the trigger file so it is in sync every 30 
	 * seconds
	*/
	pcap_dump_flush(l->pcapfd);
} 

/* insert index, offset in Judy CAM.  
 *
 * ts - string version of net/mask
 */
static void
insertJudyCAM(Word_t index, struct bnode *bn, char *ts)
{
	PPvoid_t PPvalue;
	int rc;
	Word_t offset = (Word_t) bn;

	/**************************************************/
	/* insert index, offset */
	JLI(PPvalue, Parray, index);	  // insert of index, returns value

	/* malloc failure 
	*/
	if (PPvalue == PJERR) {
		fprintf(stderr, "malloc failure with judy trees\n");
		exit(1);
	}

	*((Word_t *)PPvalue) 
			= offset;    // store of value

	/*
	printf("insert: index, size %lx %lx \n", index, *((Word_t *)PPvalue)); 
	*/
}

/* sanityCheckJudyCAM 
 *
 * basic idea:  we don't allow overlap as it is confusing
 *	as to what it might mean in terms of handling it.
 * rules: illegal if
 *	1. see if the base address matches a previous base address
 *	2. see if the base address is within a range.
 *		note: 1/2 are basically the same
 *	3. see if the base and the offset (interpreted as an address)
 *		have different predecessors.
 *
 *	returns: 1  ok
 *		 0  not ok
 */
int
sanityCheckJudyCAM(Word_t baseaddr, Word_t netmask, char *ts) 
{
	struct in_addr in; 
	Word_t indexstart = baseaddr;
	Word_t foundmask;
	Word_t netmaskstart;
	PPvoid_t PPvalue_base;
	PPvoid_t PPvalue_offset;
	int checkoffset = 0;
	struct bnode *bn;

	/* lookup baseaddr with predecessor check
	 * note: indexstart is modified if PPvalue is not null
	*/
	JLL(PPvalue_base, Parray, indexstart);	  // last index =< index passed in
	/* check baseaddr for inside region
	*/
	if (PPvalue_base != NULL) {
		bn = (struct bnode *)(*((Word_t *)PPvalue_base));
		foundmask = (Word_t) bn->netmask;
		if (((unsigned int)indexstart & (unsigned int)foundmask)  
			==  ((unsigned int)baseaddr & (unsigned int)foundmask)) {
			in.s_addr = (unsigned int) 
				htonl((unsigned int)indexstart);
			fprintf(stderr, "WARNING: addr (%s) already in previous block with base %s, new entry discarded\n", ts, inet_ntoa(in));
			return(0);
		}
	}
	/* note that it is still possible that we didn't find
	 * the baseaddress (value is NULL), we now lookup the offset
	*/
	netmaskstart = (unsigned int) 
		((unsigned int) baseaddr | (unsigned int) ~netmask);
	JLL(PPvalue_offset, Parray, netmaskstart);	  

	/*
	possibilities:
	1. PPvalue_offset is NULL,  PPvalue_base is NULL
		that's ok
	2. PPvalue_offset is not NULL, PPvalue_base is NULL
		bad as overlap
	3. PPvalue_offset is NULL, PPvalue_base is not NULL
		crazy as impossible
	4. neither are null
		ok if same base (we already know that base does
			not overlap)
	*/
	/* takes care of 1 && 4 */
	if (PPvalue_offset == PPvalue_base) {
		return(1);
	}
	in.s_addr = (unsigned int) htonl(indexstart);
	fprintf(stderr, "WARNING: discarding (%s) due to overlap with another net/mask base\n", ts);

	return(0);
}

