/*
 * hashdns.c - 
 *
 * functions:
 *	1. compute DNS metrics from DNS packets for fixed bpf graphs
 *		count:
 *		DNS requests
 *		DNS replies
 *		type A requests, replies
 *		type MX requests, replies
 *		type AAAA requests,replies 
 *		type PTR requests, replies
 *
 *	2. dns blacklist 
 *      given a blacklist of DNS names of type A, etc.
 *	print event log entries if there are hits on any of the
 *	blacklist names.
 *
 *	3. top N dns query responses. divided into non-error and error
 *		sub-reports.  
 *		include the question as well as simpler rcodes.
 *
 * config syntax:
 *	dns - turn dns facility on
 *	dns_include tag filename (list of dnsnames and types)
 *	note: multiple dns_include filenames may be given.
 *	they all go into the same hashlist. tags are used to
 *	distinguish the input file.
 * 
 */

/* bsd copyright as some of this is from tcpdump
 */
/*
 * Copyright (c) 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
 * Copyright (c) 2007 Portland State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse
 *    or promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <syslog.h>
#include <pcap.h>
#include <string.h>
#include <sys/stat.h>

#include "ourmon.h"
#include "hashdns.h"
#include "ehash.c"		/* inlined function */
#include "barthash.c"

#ifndef TRUE
#define TRUE    1
#endif
#ifndef FALSE
#define FALSE   0
#endif

static
struct dnslist dnslist;
struct dnsFlowList dnsFlowList;


static void parseDns(char *,char *, char *, int, char *);

static char *convert(char *);
static char *dnstype(int type);
static void destroy_dns_list(struct dnsFlowList *l);
static void init_dns_list(struct dnsFlowList *l);
static struct dnsFlowNode *search_flow(struct dnsFlowList *l, struct dnsFlowNode *index);
static void print_list(FILE *fd, struct dnsFlowList *l, int count);

/*
 * if dns number is in config file,
 * init dns capability.
 */
void 
enable_dns(int topn_count, char *fileName)
{
	struct dnslist *l = &dnslist;
	int x;

	//ask
	//init dnsFlowList when dns module is enabed
	init_dns_list(&dnsFlowList);

	/* put inside if (debug) {} 
	fprintf(stderr, "dns got file %s\n", fileName);
	*/

	/* means no hashlist blacklist entries
	 */
	l->blist_count = 0;
	strncpy(l->fileName, fileName, strlen(fileName));

	if (topn_count < 0 ) {
		fprintf(stderr, "dns topn_count cannot be negative %d\n", topn_count);
		exit(1);
	}
	l->topn_count = topn_count;
	l->dnsrequests = l->dnsresponses = l->acount = l->aaaacount = l->mxcount
			= l->ptr = 0;

	for (x = 0; x < 16; x++)
	{
		l->rcode_counters[x] = 0;
	}

	l->entries = 0;
}

/* insert entries into the hash list from the blacklist file
 *
 * syntax of file:
 *	dnsname  recordtype  
 *
 * record type may be A, CNAME, MX, AAAA 
 * tbd: does CNAME here make any sense?
 */


void 
config_dnslist(char *tag, char *filename)
{
	FILE *fd;
#define MAXBLISTLINE 1024
	char linebuf[MAXBLISTLINE];
	int ccount;
	char *opword;
	char *result;
	int lineNumber = 1;
	struct dnslist *l = &dnslist;
	char opbuf[1024];
	static int firsttime = 1;


	char *sltag;

	/* init hash array first time called
	 * so we can call this again to add more files with lists
	 */
	if (firsttime)
	{
		firsttime = 0;

		bzero(&l->hashlist[0], HSIZE * sizeof(struct dnsnode *));

		/* init the string list for runtime memory of actions
		 */
		initString(&l->sl);

		/*init the string list for answer queries*/
		initString(&l->slqueries);

		/*config time memory of tag strings*/

		initString(&l->sltags);
	}


	/* store the tag and later put a pointer to it in all
		 * hash entries
		 */
		sltag = storeString(&l->sltags, tag);
		if (sltag == NULL) {
			fprintf(stderr, "dns_include: ? out of malloc space for tag %s\n", tag);
			exit(1);
		}


	/* open the blacklist config file.  
	 */
	fd = fopen(filename, "r");
	if (fd == NULL)
	{
		fprintf(stderr, "dns_include open of blacklist config file %s failed\n",
		filename);
		exit(1);
	}

	/* XXX - this code is duplicated in monconfig.c, fix that somehow.
	 *
	 * while not at EOF read a line, parse it
	 */
	while (feof(fd) == 0)
	{
		result = fgets(linebuf, MAXBLISTLINE, fd);
		if (result == NULL)
			break;

		ccount = strlen(linebuf);
		linebuf[ccount-1] = 0;

		/*
		 * skip comment
		 */
		if (linebuf[0] == '#')
		{
			lineNumber++;
			continue;
		}
		/*
		 * skip empty lines
		 */
		if (linebuf[0] == '\n' || linebuf[0] == '\r')
		{
			lineNumber++;
			continue;
		}
		/*
		 * skip whitespace in line
		 */
		while (*result == ' ' || *result == '\t')
		{
			result++;
		}

		/*
		 * find first token
		 */
		strncpy(opbuf, result, 1024);
		opword = strtok(opbuf, " \t");
		if (opword == NULL || (strlen(opword) == 0))
		{
			lineNumber++;
			continue;
		}
		/*
		 * sscanf(result, "%s %s", op, value); printf("op value %s
		 * %s\n", op, value);
		 */


		(void) parseDns(filename, linebuf, result, lineNumber, sltag);
		lineNumber++;
	}
	fclose(fd);
	return;
}

/*
 * given one line from the config file of format
 *
 * dns_name_string  record_type
 * 
 * convert to internal DNS format and store in hashlist
 */


static void 
parseDns(char *filename, char *linebuf, char *result, int lineNumber, char *sltag)
{
	int rc;
	struct sockaddr_in sa;
	unsigned int ipaddr;
	struct dnsnode *n;
	struct dnsnode *h, *prev;
	unsigned int mod;
	struct dnslist *l = &dnslist;
#define MAXLINE 512
	char dc[MAXLINE] = "";
	char *token;
	char temp[MAXLINE];
	char *pt;
	char templen;
	char v1[MAXLINE] = "";
	unsigned int port;
	extern int errno;
	int tlen = 0;
	int i;
	int type;
	int len;

	temp[0] = 0;
	rc = sscanf(linebuf, "%s %s", dc, v1);
	if (rc != 2)
	{
		fprintf(stderr,
		"dns_include(%s): line %d: syntax error: dnsname type\n", filename, lineNumber);
		exit(1);
	}

	/* convert the label into the internal
	 * DNS compressed string format
	 * E.g... yakuza.com --> 6yakuza3com0 
	 */
	token = strtok(dc, ".");
	pt = temp;
	templen = 0;
	while (token != NULL)
	{
		char *tp = token;
		/* get the length of the dns component
		 */
		len = strlen(token);
		*pt++ = len; /* put the length into a byte */
		templen++; /* increment total length */
		/* copy the token into the buffer as well
		 */
		while (len--)
		{
			*pt++ = *tp++;
			templen++; /* increment total length */
		}
		token = strtok(NULL, ".");
	}
	*pt++ = 0;
	templen++;

	/* assign a certain number to each type 
	 */
	if (!strcmp(v1, "A"))
		type = 1;
	/*
	 else if(!strcmp(v1, "NS")) 
	 type = 2;
	 */
	else if (!strcmp(v1, "CNAME"))
		type = 5;
	else if (!strcmp(v1, "MX"))
		type = 15;
	else if (!strcmp(v1, "AAAA"))
		type = 28;
	else
	{
		fprintf(stderr,
		"dns_include: line %d: DNS type %s is not supported\n",
		lineNumber, v1);
		exit(1);
	}

	/* search the hash/chain list to see if dns name is already inserted
	 */
	n = l->hashlist[mod=(ehash((unsigned char*)temp, templen) % HSIZE)];
	for (; n!= NULL; n=n->next)
	{
		/* found a match, make sure if the dns name is exactly same
		 */
		/* compare dns names bcmp returns 0 if they are identical */
		/* v1 has type of the dns name such as A */
		if (n && (!bcmp(temp, n->dnsname, templen) && (type == n->type)))
		{
			return;
		}
	}

	/* note this malloc is a config-time malloc only
	 */
	n = (struct dnsnode *) malloc(sizeof(struct dnsnode));
	if (n == NULL)
	{
		fprintf(stderr, "out of space with malloc in hashdns.c\n");
		exit(1);
	}
	l->blist_count++; /* count of malloc'ed entries in blist */

	/* in addition to node you need string space
	 */
	n->next = NULL;
	n->dnsname = (char *) malloc(templen);
	if (n->dnsname == NULL)
	{
		fprintf(stderr, "out of space with malloc in hashdns.c\n");
		exit(1);
	}
	strncpy(n->dnsname, temp, templen);
	n->dnslen = templen; /* this includes the null which is petty */

	n->type = type;
	n->tag = sltag;

	/* insert the node in the hash list
	 */
	h = l->hashlist[mod];
	if (h == NULL)
	{
		l->hashlist[mod] = n;
	}
	else
	{
		n->next = h;
		l->hashlist[mod] = n;
	}
}

/* runtime: packet analysis routine
 * compute metric counters
 * analyze packet to see if hashlist hit
 */
void 
search_dns(struct t_entry *te)
{
	int rcode;
	int searched = 0;
	struct dnsnode *n;
	struct dnsFlowNode flowNodeToAdd;
	struct dns_header *np;
	unsigned char *cp;
	unsigned char i;
	int match = 0;
	struct dnslist *l = &dnslist;
	int mod;
	extern int debug;
	char *token;
	char *p;
	char *dp;
	unsigned char len;
	int notdone;
	int snaplen = te->psize;
	unsigned short qdcount, ancount, nscount, arcount;
	char *qdnsname;
	char dnsname[MAXLINE] = "";
	unsigned short tt; //16 bits
	unsigned short *ttp;
	unsigned short ans;
	unsigned short *ansp;
	char *slquery=NULL;  /*pointer to dns answer*/
	char *truequery=NULL; /*pointer to converted dns*/

	/* TCP - offset is increased by 2
	 */
	if (te->proto == IPPROTO_TCP)
		te->p += 2;
	if (snaplen <= (sizeof(struct dns_header) + 2))
		return;
	else
	/* if really too small, bail
	 */
	if (snaplen <= sizeof(struct dns_header))
		return;

	/* point to header
	 */
	np = (struct dns_header *)(te->p);

	/* get the byte-order right 
	 */
	qdcount = ntohs(np->qdcount); /* query count */
	ancount = ntohs(np->ancount); /* answer count */
	nscount = ntohs(np->nscount); /* name server count */
	arcount = ntohs(np->arcount); /* address record count */
	rcode = np->flags2 & 0x0f;

	if (debug)
	{
		printf("DNSHDR: (%s:%d->%s:%d[%d]) %x:(%x:%x:%x):%d:%d:%d:%d\n",
				intoa(te->src.s_addr), te->sport, intoa2(te->dst.s_addr),
				te->dport, te->proto, (np)->id, np->flags1, np->flags2, rcode, qdcount,
				ancount, nscount, arcount);
	}

	/* weed potential things we don't want or possible errors
	 */
	/* if response, else query, check for errors
	 */
	if ((np)->flags1 & 0x80)
	{
		if (rcode > 0)
		{ //more than 0 error?
			if (ancount > 0)
			{
				if (debug)
				{
					printf("DNSHDR: DROP qr: rcode > 0, ancount > 0\n");
				}
				return;
			}
		}
		else
		{
			/* no error and no question
			 */
			if (qdcount == 0)
			{
				if (debug)
				{
					printf("DNSHDR: DROP qr: no err & an 0\n");
				}
				return;
			}
		}
	}

	/* if query */
	else
	{
		/* rcode should not indicate an error
		 * and there should be a query count
		 */
		if (rcode || qdcount == 0)
		{
			if (debug)
			{
				printf("DNSHDR: DROP qy: rcode > 0, qdcount 0\n");
			}
			return;
		}
		/* there should not be too many answers 
		 * One observes UDP/SOA queries with qdcount 1, and ancount 1 
		 */
		if (ancount > 1)
		{
			if (debug)
			{
				printf("DNSHDR: DROP qy: ancount > 1\n");
			}
			return;
		}
		/* there should not be a name server count */
		if (nscount != 0)
		{
			if (debug)
			{
				printf("DNSHDR: DROP qy: nscount non-zero \n");
			}
			return;
		}
	}

	snaplen -= sizeof(struct dns_header);

	/* if this message is DNS query response: QR = 1 
	 * 
	 */
	if ((np)->flags1 & 0x80) { /* response */

		l->dnsresponses++;

		/* total each type of return code for responses only
		*/
		l->rcode_counters[rcode]++;

		/* if this message has an query, we record it
		 */
		if (qdcount >= 1) {

			/* skip the HEADER part including identification, flags, 
			 * the number of queries, the number of answers, 
			 * the number of authority RRs, and the number of additional RRs 
			 */
			cp = (unsigned char *)(np + 1);

			/* find query name: pointed to by qdnsname
			 */
			/* tbd check total length */
			notdone = 1;
			dp = &dnsname[0];
			if (!snaplen)
				return;
			len = *dp++ = *cp++;
			snaplen--;
			while (notdone) {
				while (len--) {
					if (!snaplen)
						return;
					*dp++ = *cp++;
					snaplen--;
				}
				if (!snaplen)
					return;
				len = *cp++ & 0xff;
				*dp++ = len;
				if (len == 0)
					notdone = 0;
			}
			qdnsname = &dnsname[0];

			/* query is in dnsname, pointed to by qdnsname
			 * cp is advanced over query
			 */

			/* compute total length of dns name */
			int tlen = strlen(qdnsname) + 1;

			/* extract Query type */
			if (snaplen < 2)
				return;
			ttp = (unsigned short *)cp;
			tt = ntohs(*ttp);

			switch (tt) {
			case 1:
				l->acount++;
				break;
			case 12:
				l->ptr++;
				break;
			case 15:
				l->mxcount++;
				break;
			case 28:
				l->aaaacount++;
				break;
			default:
				return;
			}

			/* skip query type and clall */
			cp += 4;
			snaplen -= 4;

#ifdef NOTDONE
			/*
			 * check the answer to see if it matches query 
			 * extract 2 bytes following query
			 * if the 2 bytes are equal to "c00c", 
			 * it means that the answer matches query
			 */
			if (snaplen < 2)
				return;
			ansp = (unsigned short *)cp;
			ans = ntohs(*ansp);
#endif
			truequery = convert(qdnsname);
			if (truequery == NULL) {
				if (debug) {
					printf("DNS: convert gives NULL\n"); 
				}
				return;
			}
			if (debug) {
				printf("DNS: convert gives %s\n", truequery);
			}

			/* fill in static storage flow ID
			*/

			flowNodeToAdd.dstIpAddress = (unsigned int) te->dst.s_addr;
			flowNodeToAdd.dstPort = te->dport;
			flowNodeToAdd.srcIpAddress = (unsigned int) te->src.s_addr;
			flowNodeToAdd.srcPort = te->sport;
			flowNodeToAdd.protocol = te->proto;
			flowNodeToAdd.rcode_counters = rcode;
			flowNodeToAdd.type = tt;
			flowNodeToAdd.anquery = truequery;  /* static buf */
				
			/* search for the flow id
			*/
			search_flow(&dnsFlowList, &flowNodeToAdd);

			/* we are done if no blacklist entries
			 */
			if (!l->blist_count)
				return;

			/* search hash for blacklist entries
			 */
			n = l->hashlist[mod=(ehash(qdnsname, tlen) % HSIZE)];

			if (n == NULL)
				return;

			for (; n!= NULL; n=n->next) {
				/* found a match, make sure if the dns name is exactly same 
				 */
				if (n && (!bcmp(qdnsname, n->dnsname, tlen) )) {
					match = 1;
					break;
				}
			}

			if (match) {
				extern int debug;
				unsigned int ipsrcaddr = te->src.s_addr;
				unsigned int ipdstaddr = te->dst.s_addr;
				if (debug) {
					printf("DNS BLACKLIST: tag (%s) match %d ip %x:%x q: %s\n",
							n->tag, match, ipsrcaddr, ipdstaddr, qdnsname);
				}

				l->match = 1;
				/* memorize what happened
				 */
				{
#define BUFSIZE 1024
					char buf[BUFSIZE];
					char addr1[BUFSIZE];

					int rc;
					rc = sprintf(addr1, "%s", intoa(ipsrcaddr));
					addr1[rc] = 0;
					rc = sprintf(buf, "dns query: %s (%s) from %s->%s:", 
						truequery, n->tag, 
						addr1, intoa(ipdstaddr));
					buf[rc] = 0;
					if (debug) {
						printf("DNS: storeString match: %s\n", buf);
					}
					storeString(&l->sl, buf);
				}
			}
		} /* message has query */

	} /* message is response */
	/* query */
	else
	{
		l->dnsrequests++;
	}
}

/* return a pointer to a static buffer
 * with a converted dns name
 * assume: compressed dns format.
 * return: a string with the dots in the right place.
 */
static
char * 
convert(char *internal)
{
	int l = strlen(internal);
	static char buf[1024];
	if (l == 0 || l > 256)
	{
		return (NULL);
	}

	int notdone = 1;
	char *cp = internal;
	char *dp = buf;
	unsigned char len = *cp++;
	while (notdone)
	{
		while (len--)
		{
			*dp++ = *cp++;
		}
		len = *cp++;
#define DOTCHAR 0x2e
		*dp++ = DOTCHAR;
		if (len == 0)
			notdone = 0;
	}
	*dp = 0;
	return (buf);
}

/*
 * RRDs to mon.lite include:
 *	1. dnsstats: global stats
 *	2. dnserrors0: non-error responses vs errors
 *	3. dnserrors1: fundamental errors broken down (rcode 1-6)
 *	4. dnserrors2: remaining rcodes
 */
void 
print_dns(FILE *monfd)
{
	int x = 0;
	struct dnslist *l = &dnslist;
	int rerrs = 0;
	FILE *dnsfd;

	/* TBD print global stats to mon.lite file:
	 * syntax: dnsstats: totalquery, totalresponse, Aresponse, AAAAresponse, MXresponse, CNAMErespose
	 */
	fprintf(monfd, "dnsstats: %u:%u:%u:%u:%u:%u:\n", l->dnsrequests,
			l->dnsresponses, l->acount, l->aaaacount, l->mxcount, l->ptr);

	// add up error counters
	for (x = 1; x < 16; x++)
	{
		rerrs += l->rcode_counters[x];
	}
	fprintf(monfd, "dnserrors0: %u:%u\n", l->rcode_counters[0], rerrs);
	fprintf(monfd, "dnserrors1: %u:%u:%u:%u:%u:%u:\n", l->rcode_counters[1],
			l->rcode_counters[2], l->rcode_counters[3], l->rcode_counters[4],
			l->rcode_counters[5], l->rcode_counters[6]);

	fprintf(monfd, "dnserrors2: %u:%u:%u:%u:%u:%u:\n", l->rcode_counters[7],
			l->rcode_counters[8], l->rcode_counters[9], l->rcode_counters[10],
			l->rcode_counters[11] + l->rcode_counters[12],
			l->rcode_counters[13] + l->rcode_counters[14]
					+ l->rcode_counters[15]);

	//zero dns structure
	l->dnsrequests = l->dnsresponses = l->acount = l->aaaacount = l->mxcount
			= l->ptr = 0;

	//loop through zero out rcode_counters
	for (x = 0; x < 16; x++) {
		l->rcode_counters[x] = 0;
	}

	/* if blacklist matches, log them
	 */
	if (l->match)
	{
		/* elog message if desired
		 * fprintf(monfd, "elog: IP dnslist storage has occured \n");
		 */

		dumpString(&l->sl, monfd, "elog: IP dnslist event");
		zeroString(&l->sl);
		initString(&l->sl);

		l->match = 0;
	}

	//do top N dns work: 
	//pass in a pointer where the file name is stored
	/*
	fprintf(stderr, "entering print_list %s\n", l->fileName);
	*/
	if (dnsfd = fopen(l->fileName, "w")) {
		print_list(dnsfd, &dnsFlowList, l->topn_count);
		fclose(dnsfd);
	}
	else {
		syslog(LOG_ERR, "can't open DNS file %s", l->fileName);
	}

	/* free slqueries */
	zeroString(&l->slqueries);
	initString(&l->slqueries);

}

/* called before list can be used 
 */
static
void 
init_dns_list(struct dnsFlowList *l)
{
	/*
	fprintf(stderr, "initializing flow list\n");
	*/
	l -> entries = 0;

	bzero(&l->hashlist[0], HSIZE * sizeof(struct dnsFlowNode *));
}

//comparison function to sort by total packets
static 
int sort_by_totalcount(e1, e2)
	struct dnsFlowNode *e1, *e2;
{
	if (e1->receiverHits > e2->receiverHits)
		return -1;
	if (e1->receiverHits < e2->receiverHits)
		return 1;
	return 0;
}

/*
 * search_flow
 *
 *	inputs:
 *		l - a list pointer
 *		flow as logical key
 *	returns:
 *		NULL - didn't find it
 *		struct dnsFlowNode *n 
 * side effect: if dnsFlowNode found
 */

static
struct dnsFlowNode *
search_flow(struct dnsFlowList *l, struct dnsFlowNode *index)
{
	struct dnsFlowNode *n, *h, *newNode;
	int mod;
	char *slquery = NULL;
	struct dnslist *dl = &dnslist;

	mod = barthash(index->dstIpAddress) % HSIZE;
	n = l->hashlist[mod];
	for (; n!= NULL; n=n->next) {
		// found a match

		if (n && (n->srcIpAddress == index->srcIpAddress) && (n->srcPort
				== index->srcPort) && (n->dstIpAddress == index->dstIpAddress)
				&& (n->dstPort == index->dstPort) && (n->protocol
				== index->protocol) && 
				(n->type == index->type) &&
				(n->rcode_counters == index->rcode_counters) 
			){

			//increment receiver hits
			n->receiverHits++;

			/* need to be inside if(debug)
			fprintf(stderr, "search_flow: found match\n");
			*/
			return (n);
		}
	}

	/* malloc space for question. note: done here to make sure
	 * it is only done once.
	*/
	slquery = storeString(&dl->slqueries, index->anquery);
	if (slquery == NULL) {
		syslog(LOG_ERR, "malloc failure in DNS search_flow/storeString on query - out of memory?");
		return;
	}

	/*
	fprintf(stderr, "search_flow: did not find a match\n");
	*/
	l->entries++;

	// insert the node in the hash list

	//if we hash and the linked list there has nothing then put the node there
	newNode = (struct dnsFlowNode *) malloc(sizeof(struct dnsFlowNode));

	/* check for memory failure
	*/
	if (newNode == NULL) {
		syslog(LOG_ERR, "malloc failure in DNS search_flow - out of memory?");
		return;
	}

	newNode->srcIpAddress = index->srcIpAddress;
	newNode->srcPort = index->srcPort;
	newNode->dstIpAddress = index->dstIpAddress;
	newNode->dstPort = index->dstPort;
	newNode->protocol = index->protocol;
	newNode->rcode_counters = index->rcode_counters;
	newNode->receiverHits = 1;
	newNode->type = index->type;
	newNode->anquery = slquery;

	newNode->next = NULL;

	h = l->hashlist[mod];
	if (h == NULL) {
		l->hashlist[mod] = newNode;
	}
	else {
		newNode->next = h;
		l->hashlist[mod] = newNode;
	}

#ifdef EXTREMEDEBUG
this should be a function
	struct dnsFlowNode *p;
	int i;
	for (i =0; i<HSIZE; i++)
	{
		p = l->hashlist[i];
		// if n is non-nulln use, therefore put it in the array,
		// and increment hcounter

		while (p != NULL)
		{
			fprintf(stderr, "p->dstIpAddress %u\n", p->dstIpAddress);
			fprintf(stderr, "p->srcIpAddress %u\n", p->srcIpAddress);
			fprintf(stderr, "p->dstPort %u\n", p->dstPort);
			fprintf(stderr, "p->srcPort %u\n", p->srcPort);
			fprintf(stderr, "p->protocol %u\n", p->protocol);

			p = p->next;
		}
	}
#endif
	return (NULL);
}

//------------------------ reimplement with dnsFlowNode------------
/* pack the hashlist into an array
 * and return the array.
 *
 *	returns:
 *		1 - packed array exists
 *		0 - malloc failure
 */

static struct 
dnsFlowNode *pack_array(struct dnsFlowList *l)
{
	struct dnsFlowNode *head, *n;
	unsigned int c = l->entries; // count of hash entries allocated 
	int i = 0;
	int hcounter = 0;

	/*
	fprintf(stderr, "entered pack_array\n");
	*/
	head = (struct dnsFlowNode *) malloc(c * sizeof(struct dnsFlowNode));
	if (head == NULL) {
		//fprintf(stderr, "pack_array, malloc failed\n");
		syslog(LOG_ERR, "malloc failure in port/pack_array, count %d",
			l->entries);
		return (NULL);
	}
	// walk the hashlist

	for (i =0; i<HSIZE; i++)
	{
		n = l->hashlist[i];
		// if n is non-nulln use, therefore put it in the array,
		// and increment hcounter

		while (n != NULL)
		{
			/*
			fprintf(stderr, "n->dstIpAddress %u\n", n->dstIpAddress);
			fprintf(stderr, "n->srcIpAddress %u\n", n->srcIpAddress);
			fprintf(stderr, "n->dstPort %u\n", n->dstPort);
			fprintf(stderr, "n->srcPort %u\n", n->srcPort);
			fprintf(stderr, "n->protocol %u\n", n->protocol);
			*/
						
			head[hcounter].srcIpAddress = n->srcIpAddress;
			head[hcounter].srcPort = n->srcPort;
			head[hcounter].dstIpAddress = n->dstIpAddress;
			head[hcounter].dstPort = n->dstPort;
			head[hcounter].protocol = n->protocol;
			head[hcounter].receiverHits = n->receiverHits;
			head[hcounter].type = n->type;
			head[hcounter].rcode_counters = n->rcode_counters;
			head[hcounter].anquery = n->anquery;

			n = n->next;
			hcounter++;
		}
	}

	return (head);
}

/* given a udplist, sort it,
 * and print topN udp results.
 *
 * inputs:
 *      fd - file for print output
 *      l - hash array
 *      tcpflag/udpflags - flags to indicate various filter outputs
 *      count - max desired outputs for all filters
 */

static
void 
print_list(FILE *fd, struct dnsFlowList *l, int topn_count)
{
	struct dnsFlowNode *head;
	int i;
	int c = l->entries;   /* total number of entries */
	int consumed;

	/*
	fprintf(stderr, "entered print_list\n");
    */

	// no of hash elements
	// form the hashlist into a packed array

	head = pack_array(l);
	if (head == NULL) {
		/* TBD. syslog wouldn't hurt.
		fprintf(fd, "malloc error in print_udpplist\n");
		*/
		return;
	}

	/* quick sort by total flow count
	*/
	qsort((void *)head, c, sizeof(struct dnsFlowNode), sort_by_totalcount);


	fprintf(fd, "# dns response topn flow info (no errors): topn_count: %u, total flow entries: %u\n", 
			topn_count, l->entries);

	consumed = 0;	/* the count of items consumed against the topn supplied value */
	/* print non error entries
	*/
	for (i = 0; i < c; i++)
	{
		/* if exceeding total topn request count, stop
		*/
		if (topn_count > 0 && consumed >= topn_count) {
			break;
		}
		/* don't exceed total number of entries either
		*/
		if (consumed >= c) {
			break;
		}

		if (!head[i].rcode_counters) {
			fprintf(fd, "%s.%u->%s.%u:q:%s:hits:%u:type:%s:errs:%d\n",
				intoa(head[i].srcIpAddress), head[i].srcPort, 
				intoa2(head[i].dstIpAddress), head[i].dstPort, head[i].anquery,
				head[i].receiverHits, dnstype(head[i].type), head[i].rcode_counters);
			consumed++;
		}

	}
	fprintf(fd, "\n");


	/*
	 * print info sorted by error count. given that error count is part
	 * of the flow tuple, we just find the top N error flows
	 */

	fprintf(fd, "# rcode topn errors\n");
	i = 0;		/* the index into the head array */
	consumed = 0;	/* the count of items consumed against the topn supplied value */
	for (i = 0; i < c; i++) {
		if (topn_count > 0 && consumed >= topn_count) {
			break;
		}
		if (consumed >= c) {
			break;
		}

		/* if an error tuple */
		if (head[i].rcode_counters) {
			fprintf(fd, "%s.%u->%s.%u:q:%s:hits:%u:errs:%d\n",
					intoa(head[i].srcIpAddress), head[i].srcPort, 
					intoa2(head[i].dstIpAddress), head[i].dstPort, head[i].anquery,
				    head[i].receiverHits, head[i].rcode_counters);

			consumed++;
		}
	}
	fprintf(fd, "\n");

	free(head);

	destroy_dns_list(l);
	init_dns_list(l);

}

/*
 * deallocate malloced hash entries
 */

static void 
destroy_dns_list(struct dnsFlowList *l)
{
	int i;
	struct dnsFlowNode *n;
	struct dnsFlowNode *prev;
	/*
	fprintf(stderr, "destroying list\n");
	*/
	for (i =0; i< HSIZE; i++) {
		n = l->hashlist[i];
		while (n != NULL)
		{
			prev = n;
			n = n->next;
			free(prev);
		}
	}
}

static char *
dnstype(int type)
{
	static char buf[1024];
	switch (type) {
	case 1:
		return("A");
		break;
	case 12:
		return("PTR");
		break;
	case 15:
		return("MX");
		break;
	case 28:
		return("AAAA");
		break;
	default:
		sprintf(buf, "%d", type); 
		return(buf);
	}
}
