#! /usr/bin/perl

#
# ssh.pl - summarize ssh activity
#
# TODO: Document this script. 
#
# may be run:
#	per sample
#	per N samples (e.g., all ssh sample files in a daily log directory)
#
# For syntax:
#	# ssh.pl -d <directory> | -f "rawssh.txt0 ... rawssh.txtN"
#	    [-h 10.0.0.0/8 ] [ -debug ] [-n] > stdout
#
# this is rough:  See the function printUsageAndExit for complete
# syntax.  ssh.pl is also self-prompting.
#
# input file(s):
#	ssh.txt raw input file. 
#
# outputs:	   
#	stdout
#
# events:
#   TODO: determine what's getting into the event log.
#
# author/s:
#	amorton@pdx.edu
#
# BUGS:
#
#   ?
#
########################################################

use strict;
use Socket;
use Time::Local;
use File::Basename;
use vars qw/ %opt /;
use Getopt::Long;
use Data::Dumper;
use List::Util qw[min max];

startTimer();


# Constants that affect the output.
#
# The number of seconds a single file represents.
my $FILE_TIMESPAN = 30;
# Display hosts with this many opened connections.
my $TOTAL_SYNS;
# Display hosts with this many reset connections
my $TOTAL_RSTS;
# When displaying lists of connections display the top N hosts.
my $SUBLIST = 3;
# Number of connections to the same server to get you flagged for password
# guessing.
my $BRUTE;
# Number of hosts to connect to before you get flagged for scanning.
my $SCANNING;
# Maximum L3D fanout before you get flagged for scanning.
my $MAX_L3D = 50;
# maximum number of evil connected flows observed before we log event
my $MAX_BAD_CONNS = 2;
# minimum average size to be considered scp instead of password guessing
my $MIN_AVG_PKT_SIZE = 250;
# maximum ratio of sf flows to total flows to be considered normal traffic instead of scp or password guessing
my $MAX_SYN_FIN_RATIO = 0.74;

# We connect to cymru for country code lookups and don't want to kill them in
# case of DOS attack.
my $LOOKUP_LIMIT = 500;	# Maximum number of DNS lookups.
my $lookups = 0;		# Running total of DNS lookups.

GetOptions(
	"help" => \$opt{help},
	"debug" => \$opt{debug},
	"h=s" => \$opt{h},
	"b=s" => \$opt{b},
	"c=s" => \$opt{c},
	"d=s" => \$opt{d},
	"e:s" => \$opt{e},
	"f:s" => \$opt{f},
	"z:s" => \$opt{z},
	"k:s" => \$opt{k},
	"n" => \$opt{n},
);

my $DEBUG = 0;
if ( $opt{'debug'} ) {
	$DEBUG = 1;
}

my %EvilIPs; 
my @temp;
if ( $opt{'z'} ) {
	dbmopen(%EvilIPs, $opt{'z'}, 0666) or
		die "can't open blacklist db file $opt{'Z'}: $1\n";
}

my @FILES;
if ( $opt{'d'} ) {
	my $directory = $opt{'d'};

	$TOTAL_SYNS = 1;
	$BRUTE = 10;
	$TOTAL_RSTS = $SCANNING = 5;

	#
	# read all files in a directory into the array and
	# prepend directory path to filename
	#
	opendir (DIR, $directory) or 
		die "can't open $directory directory: $!\n";
	@FILES = readdir DIR;
	closedir DIR;
	my $size = scalar @FILES;
	while ($size > -1) {
		# prepend directory path to files
		my $f = $directory.'/'.$FILES[$size];
		if ( -f $f && ! -d $f ) {
			$FILES[$size] = $f;
		}
		$size--;
	}
}
elsif ( $opt{'f'} ) {
	$TOTAL_SYNS = $TOTAL_RSTS = $BRUTE = $SCANNING = 1;
	@FILES = split(/ /, $opt{'f'});
}


# Print the help an exit if they requested it or if we don't have a file.
my $fileCount = scalar @FILES;
if ( $fileCount < 1 || $opt{'help'} ) { 
	printUsageAndExit(); 
}

my $eventFile = "";
my $eventFlag = 0;
if ( $opt{'e'} ) {
	$eventFile = $opt{'e'};
	$eventFlag = 1;
}

# filter output based on a net/mask pair.  i.e., ip must be "inside"
# net/mask.
my $homeNetmask;
my $homeNetmaskFlag = 0;
if ( $opt{h} ) {
	$homeNetmask = $opt{h};	# netmask address should be 10/8 or the like.
	$homeNetmaskFlag = 1;		# capability is on
	if ($DEBUG) {
		print STDERR "Using $homeNetmask as the home network\n";
	}
	
}

my $darkNetmask;
my $darkNetmaskFlag = 0;
if ( $opt{k} ) 
{
	$darkNetmask = $opt{k};
	$darkNetmaskFlag = 1;
	if ($DEBUG) {
		print STDERR "Using $darkNetmask as the darknet\n";
	}
}

my %evilCC;
if ( $opt{'b'} ) {
	@evilCC{split(/,/, uc($opt{'b'}))}++; 
	
	if ($DEBUG) {
		print STDERR 'Looking for these "evil" countries: '. join(', ', keys %evilCC) ."\n";
	}
}

# Mapping of IP -> two letter country code.
my %countryByIp;
my $dbFile = '';
if ( $opt{'c'} ) { 
	$dbFile = $opt{'c'};
	if ($DEBUG) { 
		print STDERR "Opening db cache file: $dbFile\n"; 
	}
	dbmopen(%countryByIp, $dbFile, 0666);
}

# Total count of RSTs
my %rstsRecv;
# RSTs to a given IP
my %rstsByDest;

# Total counts of SYNs sent and received.
my %synsSent;
my %synsRecv;

# Total counts of packets sent and received.
my %pktsSent;
my %pktsRecv;

# Total count of bytes sent and received.
my %bytesSent;
my %bytesRecv;

# Counts of SYNs by sender and receiver.
my %synsBySrc;
my %synsByDest;

# Total counts of FINs 
my %finsSent;
my %finsRecv;
my %finsBySrc;
my %finsByDest;

# syn src flows - src by instance, value is flow count
my %synSrcInstances;
# similar to above but all flows, src is index 
my %flowSrcInstances;
# similar to above but flows sans s/f/reset
my %flowNoSYNFINInstances;
# traces an approximation of an active connection, resets to 0 if interrupted
my %flowNoSYNFINPairs;
my %lastNoSYNFINPair;
my %consecNoSYNFINInstances;
# flows with only a syn, no fin
my %flowSOInstances;
#  flows with SF only (resets ignored)
my %flowSFInstances;
# any flow src by destination -- we count the total flows
my %flowAnySrcDest;

my $fileindex = 0;
# Work through the list of files.
while (my $file = shift @FILES ) {
	if (! -f $file) {
		next;
	}

	open(FILE, $file) || die "can't open $file file: $!\n";
	my @lines = <FILE>;
	if ($DEBUG) {
		print STDERR "Reading $file (". scalar(@lines) ." lines)\n";
	}
	close(FILE);

	$fileindex++;

	# Convert the filename ("Feb__7_06:00:01_PST_2008.ssh.txt") into a timestamp.
	my $time = "";
	if (basename($file) =~ /(\w*)_(\d*)_(\d*):(\d*):(\d*)_(\w*)_(\d*)/ )
	{
		$time = timelocal($5, $4, $3, $2, $1, $7);
		if ($DEBUG) { print STDERR "Timestamp is $time\n"; }
	}

	# process on line by line basis
	foreach my $line (@lines) {
		# get rid of blanks, return, line breaks
		$line =~ s/\n//g;
		$line =~ s/ //g;

		# Extract hosts, ports, and syn, fin and rst counts from:
		# topn flow tuple
		# 12.48.36.7.51616->131.252.212.244.22:[6]:148:2:2:0:0:0:0:0:0:0:2:0:0:
		# note anchored at the end as c1:c2:c3 are at the end.
		if ($line =~ /(.*)\.(\d*)->(.*)\.(\d*):\[.*]:(\d*):(\d*).*:(\d*):(\d*):(\d*):$/ ) {
			my ($src, $srcp, $dest, $destp, $bytes, $packets, $syns, $fins, $rsts)  = ($1, $2, $3, $4, $5, $6, $7, $8, $9);
#			print STDERR "DEBUG! bytes=$bytes  packets=$packets \n";

			# Who's trying to connect to machines without SSH running?
			# src port 22: possible server and generating resets to dest
			# note: resets are from src to dst (as are syns/fins)
			
			if ($srcp == 22 && $rsts > 0) {
				# tag offender with resets
				$rstsRecv{$dest} += $rsts;

				# ... and the counts by target.
				# TBD. not used.
				$rstsByDest{$dest}{$src} += $rsts;
			}

			# server port 22, therefore client
			if ($destp == 22 && $syns > 0) {
				# Keep a total count of opened connections for each end...
				$synsSent{$src} += $syns;
				$synsRecv{$dest} += $syns;

				# ... who's calling how many times ...
				$synsBySrc{$src}{$dest} += $syns;

				# ... and who's being called by who how many times.
				$synsByDest{$dest}{$src} += $syns;

				# this is really a flow count per instance
				# for syn sender
				$synSrcInstances{$src}->{$fileindex} += 1;
			}

			# Look at connections being closed (FINs) on port 22.
			if ($fins > 0) {
				# Keep a total count of closed connections for each end...
				$finsSent{$src} += $fins;
				$finsRecv{$dest} += $fins;

				# ... who's calling how many times ...
				# TBD. not used
				# $finsBySrc{$src}{$dest} += $fins;

				# ... and who's being called by who how many times.
				# TBD. not used
				# $finsByDest{$dest}{$src} += $fins;
			}

			# flow counts:
			$flowAnySrcDest{$src}{$dest} += 1;

			# all flows by src only
			$flowSrcInstances{$src} += 1;

			# global packet counts
			$pktsSent{$src} += $packets;
			$pktsRecv{$dest} += $packets;
			$bytesSent{$src} += $bytes;
			$bytesRecv{$dest} += $bytes;

			# no SF and no resets either
			# assumption: approximation of connected
			if ($syns == 0 && $fins == 0 && $rsts == 0) {
				$flowNoSYNFINInstances{$src} += 1;
				if ($time ne "")
				{
					if (($flowNoSYNFINPairs{$src}{$dest} == 0) || ($lastNoSYNFINPair{$src}{$dest} + 30 < $time))
					{
						$flowNoSYNFINPairs{$src}{$dest} = 1;
						$lastNoSYNFINPair{$src}{$dest} = $time;
					}
					elsif ($lastNoSYNFINPair{$src}{$dest} + 30 >= $time)
					{
						$flowNoSYNFINPairs{$src}{$dest}++;
						$lastNoSYNFINPair{$src}{$dest} = $time;
					}
			 		if ($flowNoSYNFINPairs{$src}{$dest} > $consecNoSYNFINInstances{$src}{$dest})
					{
						$consecNoSYNFINInstances{$src}{$dest} = $flowNoSYNFINPairs{$src}{$dest};
					}
				}
			}
			# S only
			if ($syns > 0 && $fins == 0) {
				$flowSOInstances{$src} += 1;
			}
			# SF both
			# note: this means either quickie data xfer OR
			# password guessing with failures
			if ($syns > 0 && $fins > 0) {
				$flowSFInstances{$src} += 1;
			}
		}
	}  # end lines loop
} # end files loop


# Lookup the country codes for IP addresses we've encountered.
if ( ! $opt{'n'} ) {
	lookupCountries( (keys %synsSent, keys %synsRecv, keys %rstsRecv, keys %finsSent, keys %finsRecv ) );
}


print "# Top sources of opened SSH connections (with more than $TOTAL_SYNS SYNs)\n";
print "# total 30-second samples: $fileCount\n";
#print "# format: ip cc instances L3D total_syn/fin/reset flows/total/sf/so/nosf\n";
printSynHeader();
my $fins; 
my $rsts;
my @srcBySynsSent = sort { $synsSent{$b} <=> $synsSent{$a} } keys %synsSent;
foreach my $src ( @srcBySynsSent ) {
	if ($synsSent{$src} >= $TOTAL_SYNS) {
		checkBlacklist($src, $synsSent{$src} ." SYNs");

		my $cc = $countryByIp{$src} ? "$countryByIp{$src}" : "??";
		my $fins = $finsSent{$src};
		if (not defined $fins) {
			$fins = 0;
		}
		my $rsts = $rstsRecv{$src};
		if (not defined $rsts) {
			$rsts = 0;
		}
		my $ic = 0;
		foreach my $instance (keys %{$synSrcInstances{$src}}) {
			$ic++;
		}
		# flow stats
		my $tf = $flowSrcInstances{$src};
		if (not defined $tf) {
			$tf = 0;
		}
		my $so = $flowSOInstances{$src};
		if (not defined $so) {
			$so = 0;
		}
		my $sf = $flowSFInstances{$src};
		if (not defined $sf) {
			$sf = 0;
		}
		my $co = $flowNoSYNFINInstances{$src};
		if (not defined $co) {
			$co = 0;
		}
 	 
		# determine L3D stat 
		my $L3D = 0;
       		foreach my $mydst (keys %{ $flowAnySrcDest{$src} }) {
			$L3D++;
       		}

		# determine average packet size
		my $sizer = int ( $bytesSent{$src} / $pktsSent{$src} );

		my $guess = ""; # don't know anything, don't say anything
		if ($EvilIPs{$src} =~ /.*s.*/)
		{
			$guess = $guess."s";
		}
		if ($EvilIPs{$src} =~ /.*d.*/)
		{
			$guess = $guess."d";
		}
		if ($rsts > $SCANNING)
		{
			$guess = $guess."r"; 
		}
		if ($L3D > $MAX_L3D)
		{
		 	$guess = $guess."L"; 
		}
		if (($sf+$so)/$tf > $MAX_SYN_FIN_RATIO)
		{
			$guess = $guess."W";
			if (($sf/$tf) > $MAX_SYN_FIN_RATIO)
			{
				$guess = $guess."F";
			} 
			if ($sizer < $MIN_AVG_PKT_SIZE)
			{
			# small packets are OK if F flag is not triggered	
				$guess = $guess."P";
			}
		}
		foreach (keys %{$synsBySrc{$src}})
		{
			if (insubnet($darkNetmask, $_))
			{
				$guess = $guess."D"; 
				if ($EvilIPs{$src} ne "")
				{
					$EvilIPs{$src} = $cc."d"; 
				}
				elsif ($EvilIPs{$src} !~ /.*d.*/)
				{
					$EvilIPs{$src} = $EvilIPs{$src}."d"; 
				}
				last;
			}
		}
		if ($EvilIPs{$src} ne "")
		{
			foreach (keys %{$consecNoSYNFINInstances{$src}})
			{
				if ($consecNoSYNFINInstances{$src}{$_} > $MAX_BAD_CONNS)
				{
					$guess = $guess."C";
				}
			}
		}
	 	if (($guess =~ /.*D.*/) and ($countryByIp{$src} eq "home"))
		{
			storeEvent($eventFile, "home IP $src ssh scanning into local darknet");
		}	
		if (($guess =~ /.*s.*/) and ($countryByIp{$src} eq "home"))
		{
			storeEvent($eventFile, "home IP $src ssh scanning based on static ssh blacklist");
		}

		my $sbuf = sprintf("%-16.16s %-6.6s %-8.8s %-12.12s %-16.16s %-26.26s %s\n",
		"$src",
		"$cc",
		"$guess",
		"$ic", 
		"$L3D",
		"$synsSent{$src}:$fins:$rsts",
		"$tf:$sf:$so:$co:$sizer"
		);
		print "$sbuf";

		# TBD. format this and header
		#print "$src\t$cc\t $ic\t $L3D\t $synsSent{$src}/$fins/$rsts\t$tf:$sf:$so:$co\n";
	}
}

print "\n# Key to the guess field codes\n";
print "s = static blacklist hit\n";
print "d = dynamic blacklist hit\n";
print "D = sending to darknet\n";
print "r = too many resets received\n";
print "F = too many sf flows\n";
print "W = too many sf+so flows\n";
print "P = packets are too small to be scp\n";
print "C = more than $MAX_BAD_CONNS in a row\n";
print "L = too broad a destination fan-out\n";

# reset section
##############################################################################################
print "\n# Top connecters to non-existant SSH servers (with more than $TOTAL_RSTS RSTs)\n";

my @srcByRstsSent = sort { $rstsRecv{$b} <=> $rstsRecv{$a} } keys %rstsRecv;

foreach my $src ( @srcByRstsSent ) {

	if ($rstsRecv{$src} >= $TOTAL_RSTS) {
		checkBlacklist($src, $rstsRecv{$src} ." RSTs");
		my $cc = $countryByIp{$src} ? "\t$countryByIp{$src}" : '';
		print "$src$cc\t$rstsRecv{$src} rsts\n";
	}
}

print "\n# Scanning: Source IP with targets and number of connections (Only showing $SCANNING or more targets).\n";
my @srcByTotalSyns = sort { keys %{$synsBySrc{$b}} <=> %{$synsBySrc{$a}} } keys %synsBySrc;
foreach my $src ( @srcByTotalSyns ) {
	# Print any one who's talking to more than $SCANNING hosts.
	my @dests = sort { $synsBySrc{$src}{$b} <=> $synsBySrc{$src}{$a} } keys %{$synsBySrc{$src}};
	my $count = scalar(@dests);
	if ($count >= $SCANNING) {
		checkBlacklist($src, " SYNs to $count targets");
		my $cc = $countryByIp{$src} ? "\t$countryByIp{$src}" : '';
		print "$src$cc\t-> $count targets: ";
		my $limit = ($count > $SUBLIST ? $SUBLIST : $count) - 1;
		foreach my $dest (@dests[0..$limit]) {
			my $cc = $countryByIp{$dest} ? "$countryByIp{$dest}, " : '';
			print "$dest ($cc$synsBySrc{$src}{$dest} syns) ";
		}
		if ($count > $limit + 1) {
			print "...";
		}
		print "\n";
	}
}



print "\n# Possible password guessing: Targeted IP with senders and number of connections (Only showing $BRUTE or more connections).\n";
my @destByTotalSyns = sort { $synsRecv{$b} <=> $synsRecv{$a} } keys %synsRecv;
foreach my $dest ( @destByTotalSyns ) {
	my @srcs = sort { $synsByDest{$dest}{$b} <=> $synsByDest{$dest}{$a} } keys %{$synsByDest{$dest}};
	my $count = scalar(@srcs);
	# Since we're trying to find brute force attacks ignore targets that hit less than $BRUTE times.
	if ($synsByDest{$dest}{@srcs[0]} >= $BRUTE) {
		my $cc = $countryByIp{$dest} ? "\t$countryByIp{$dest}" : '';
		print "$dest$cc\t<- ";
		for (my $i = 0; $i < $count && $synsByDest{$dest}{@srcs[$i]} >= $BRUTE; $i++) {
			my $cc = $countryByIp{@srcs[$i]} ? "$countryByIp{@srcs[$i]}, " : '';
			print "@srcs[$i] ($cc$synsByDest{$dest}{@srcs[$i]} syns) ";
		}
		print "\n";
	}
}


#print "\n# Most highly connected hosts \n";
#my @mostConnected = sort { keys %{$consecNoSYNFINInstances{$b}} <=> keys %{$consecNoSYNFINInstances{$a}} } keys %consecNoSYNFINInstances;
#my $n = @mostConnected; 
#foreach my $src ( @mostConnected ) 
#for (my $i = 0; $i < 10 && $i < $n; $i++)
#{
#	my $src = $mostConnected[$i];
#	my $nn = keys(%{$consecNoSYNFINInstances{$src}}); 
#	my @connDests = sort { $consecNoSYNFINInstances{$src}{$b} <=> $consecNoSYNFINInstances{$src}{$a} } keys %{$consecNoSYNFINInstances{$src}};
#	if ($consecNoSYNFINInstances{$src}{$connDests[0]} >= 10)
#	{
#		print "host $src has $n connections (top 5 connections shown)\n";
#		for (my $ii = 0; $ii < 5 && $ii < $nn; $ii++)
#		{
#			print $consecNoSYNFINInstances{$src}{$connDests[$ii]}." continguous no-syn/no-fin flows with ".$connDests[$ii]."\n";
#		}
#	}
#}

#foreach my $src (keys %synSrcInstances) 
#{
#	my $ic = 0;
#	my $tf = 0;	# total syn flows
#	foreach my $instance (keys %{$synSrcInstances{$src}}) {
#		$ic++;
#		$tf += $synSrcInstances{$src}->{$instance};
#		print "XXX $src: $instance: $synSrcInstances{$src}->{$instance}\n";
#		
#	}
#	print "src instance: $src, $ic, $tf\n";
#}

#print "\n# Ratio of closed SSH server connections--FINs to files ($fileCount)\n";
# sort by fins sent
#my @srcByFinsSent = sort { $finsSent{$b} <=> $finsSent{$a} } keys %finsSent;
#foreach my $src ( @srcByFinsSent ) {
#	my $ratio = $finsSent{$src} / $fileCount;
#	if ($ratio >= 0.5) {
#		my $cc = $countryByIp{$src} ? "\t". $countryByIp{$src} : '';
#		printf "%s%s\t%.4f fins/file\n", $src, $cc, $ratio;
#	}
#}

#print "\n# Ratio of closed SSH client connections--FINs to files ($fileCount)\n";
#my @destByFinsRecv = sort { $finsRecv{$b} <=> $finsRecv{$a} } keys %finsRecv;
#foreach my $dest ( @destByFinsRecv ) {
#	my $ratio = $finsRecv{$dest} / $fileCount; 
#	if ($ratio >= 0.5) {
#		my $cc = $countryByIp{$dest} ? "\t". $countryByIp{$dest} : '';
#		printf "%s%s\t%.4f fins/file\n", $dest, $cc, $ratio; 
#	}
#}


if ($dbFile) {
	if ($DEBUG) { print STDERR "Closing db cache file.\n"; }
	dbmclose(%countryByIp);
}


my $secs = getElapsedTime();
print "\nTotal time: $secs seconds: Total dig calls: $lookups of $LOOKUP_LIMIT.\n";  

exit;


########################################################## 

#
# subroutines
#
sub printUsageAndExit
{
	print "Usage: $0 -d directory | -f \"file1 file2 ... fileN\"\n";
	print "    -debug              display debugging messages\n";
	print "    -help               displays usage\n";
	print "    -b CC1,CC2,CC3      black list of country codes\n";
	print "    -c cacheFile        caches host's info: country code, first seen, etc.\n";
	print "    -d directory        directory of mon.lite files\n";
	print "    -e eventfile        event file for event logging\n";
	print "    -f file1 file2 ...  fileN list of files to read\n";
	print "    -h network/mask     home network/mask (e.g., 10.0.0.0/8)\n";
	print "    -k network/mask     darknet network/mask\n";	
	print "    -n                  do not look up country codes\n";
	print "    -z evilmap          a db map file of static or dynamically determined evil IPs\n";
	exit;
}

#
# Check if an IP is in the list of "evil" countries.
#
# checkBlacklist($ip, $trafficType)
#
sub checkBlacklist
{
	my $ip = $_[0];
	my $trafficType = $_[1];
	my $cc = $countryByIp{$ip};
	
	if (exists $evilCC{$countryByIp{$ip}}) {
		storeEvent($eventFile, "SSH traffic ($trafficType) with  blacklisted country: ", "$countryByIp{$ip} $ip ");
	}
}

#
# Look up the network that the connection origininated from.
#
sub lookupCountries
{
	foreach my $ip ( @_ ) {
		if ($DEBUG) { print STDERR "lookupCountries: --> $ip\n"; }
 
		if ($countryByIp{$ip}) {
			# Skip it if it's already in the cache.
			if ($DEBUG) { print STDERR "Cache hit $ip\t-> $countryByIp{$ip}\n"; }
		}
		# Don't lookup home network IPs.
		elsif ($homeNetmaskFlag && insubnet($homeNetmask, $ip)) {
			$countryByIp{$ip} = 'home';
			if ($DEBUG) { print STDERR "Skipped $ip\t-> home network\n"; }
		}
		# Don't lookup private IPs.
		elsif (insubnet('10/8', $ip) || insubnet('172.16/12', $ip) || insubnet('192.168/16', $ip)) {
			$countryByIp{$ip} = 'private';
			if ($DEBUG) { print STDERR "Skipped $ip\t-> private network\n"; }
		}
		# We don't want to kill Cymru with lookups in case of a DDOS.
		elsif ($lookups >= $LOOKUP_LIMIT) {
			if ($lookups == $LOOKUP_LIMIT) {
				storeEvent($eventFile, 'SSH country code lookup exceeded the limit: ', $LOOKUP_LIMIT);
			}
		}
		else {
			# Query the Cymru system for a country code.
			my $query = join('.', reverse(split(/\./, $ip))) .'.origin.asn.cymru.com';
			if ($DEBUG) { print STDERR "params to dig +short $query TXT"; }
			my $result = `dig +short $query TXT`;
			chomp($result);
			$lookups++;

			# The results should have one or more lines like this:
			#   "6366 | 131.252.0.0/16 | US | arin | 1989-02-13"\n
			# we'll just use the first one.
			my @lines = split("\n", $result);
			if (scalar @lines) {
				# Trim off the quotes then split it up by the pipes.
				$lines[0] =~ s/\"//g;
				my @parts = split(/ \| /, $lines[0]);
				if ($DEBUG) { print STDERR "Looked up $ip\t-> $parts[2]\n"; }
				$countryByIp{$ip} = $parts[2];
			}
			else {
				$countryByIp{$ip} = '??';
				if ($DEBUG) { print STDERR "Error looking up $ip.\n"; }
			}
		}
	}
}

# Variables used for tracking elapsed time.
my $startTime;
my $tstart = 0;

#
# Start the timer.
#
sub startTimer
{
	$startTime = `date`;
	$startTime =~ s/\n//g;
	$tstart = time();
}

#
# Compute the elapsed time.
#
sub getElapsedTime
{
	return time() - $tstart;
}


# storeEvent($eventFile, "new worm signature from:", $sbuf);
#
# storeEvent(file, prependString, eventString)
#
# eventString may or may not have a newline.  we strip it if it does.
#
sub storeEvent
{
	if ($eventFlag == 0) {
		return;
	}
	my $eventFile = $_[0];
	my $prependString = $_[1];
	my $eventString = $_[2];
	my $s;

	$eventString=~ s/\n//g;
	$s = "$startTime: " . $prependString . $eventString;

 	open(FILE, ">>$eventFile") || die "can't open $eventFile: $!\n";
 	print FILE "$s\n";
 	close(FILE);
}

# insubnet
#
# note: input param netmask may have these forms
#	1.0.0.0/8
#	1.0.0.0/8,2.0.0.0/8
#	1/8
#	10.10/16
#	1/8,2/8,200.1.2.0/24
#
# a comma-separated list is possible.
# so called netmask variable below is really net/mask.
#
# compare a subnet/mask to an ip
# return:
#	1 if ip within subnet/s
#	0 otherwise

sub insubnet
{
	my $netmask = $_[0];
	my $ip = $_[1];
	my $hexmask;
	my $rnet1;
	my $rnet2;
	my $net1;
	my $net2;
	my $index=0;
	
	my @nets = split(",", $netmask);

	while ($nets[$index]) {

		my @words = split("\/", $nets[$index]);
		my $subnet = $words[0];
		my $mask = $words[1];

		# fix subnet if too short - we need full dot quad notation
		my @quads=split('\.',$subnet);
		my $suffixquads = 3 - $#quads;  # number of .zero quads needed at end
		# append enough dot quads if needed
		if ($suffixquads == 3) {
			$subnet = $subnet . ".0.0.0";
		}
		elsif ($suffixquads == 2) {
			$subnet = $subnet . ".0.0";
		}
		elsif ($suffixquads == 1) {
			$subnet = $subnet . ".0";
		}

		# if slash notation
		# 1. convert the specified mask
		# to a hex value (a real mask)
		$hexmask = 0xffffffff << (32 - $mask);
		# limit to 32 bits please
		$hexmask = $hexmask & 0xffffffff;

		# ip address based netmask as opposed to slash
		#$hexmask = unpack('N', inet_aton($mask));

		# 2. convert the ipaddr to a host int
		$net1 = unpack('N', inet_aton($ip));
		$net2 = unpack('N', inet_aton($subnet));
		# printf ("inputip: %x inputnet %x mask %x\n", $net1, $net2, $hexmask);

		# 3. compare the results
		$rnet1 = $net1 & $hexmask;
		$rnet2 = $net2 & $hexmask;

		# print "$rnet1, $rnet2:", "$rnet1 == $rnet2\n"; 
		if ($rnet1 == $rnet2) {
			return (1);
		}
		$index++;
	}
	return(0);
}

sub printSynHeader()
{

	#my $sbuf = sprintf("%-16.16s %-16.16s %-12.12s %-16.16s %8s: %8s: %8s: %9s:
	my $sbuf = sprintf("%-16.16s %-6.6s %-8.8s %-12.12s %-16.16s %-26.26s %s\n",
                "ip src:",
                "CC:",
		"guess:",
                "samples#:",
                "L3D:",
                "syn/fin/reset:",
                "tf/sf/so/nosf/size");
        print "$sbuf";
}
